
        ################################################################
        #                                                              #
        #                    Python FITS Object Class                  #  
        #                                                              #
        #                           D. Kelson                          #  
        #                          6 May  2001                         #  
        #                                                              #
        ################################################################

########################################################################

from types import *
from string import *
from os import uname, path, remove
from sys import exit
from re import sub
from MLab import squeeze
from Numeric import *
import CFITSIO2 as C
import tempfile

########################################################################

bitpixes =   [8,   16, 16,    32,   32,   -32    ,   -32]
datatypes=   [Int8, "b", "s" ,"i" ,"l" ,"f","d"]
rfunctions=[C.read_subset_int, C.read_subset_int, C.read_subset_int, C.read_subset_int,
        C.read_subset_lng, C.read_subset_flt, C.read_subset_dbl]
wfunctions=[C.write_subset_int, C.write_subset_sht, C.write_subset_sht, C.write_subset_int,
        C.write_subset_lng, C.write_subset_flt, C.write_subset_dbl]

########################################################################

class FITS:

  def __init__(self,fitsfile,readonly=1,create=0,template=None):
        actfile = split(fitsfile,'[')[0]
        if create and template:
           if path.exists(fitsfile): remove(fitsfile)
           status,fp=C.create_template(fitsfile,template)
        elif create:
           if path.exists(fitsfile): remove(fitsfile)
           status,fp=C.create_file(fitsfile)
        elif readonly and path.exists(actfile):
           if not path.exists(actfile): raise "%s does not exist!" % (actfile)
           status,fp=C.open_file(fitsfile,C.READONLY)
        elif readonly and not path.exists(actfile) \
        and path.exists( replace(fitsfile, actfile, actfile+".gz")):
           fitsfile = replace(fitsfile, actfile, actfile+".gz")
           if not path.exists(fitsfile): raise "%s does not exist!" % (fitsfile)
           status,fp=C.open_file(fitsfile,C.READONLY)
        elif not readonly:
           if not path.exists(fitsfile): raise "%s does not exist!" % (fitsfile)
           status,fp=C.open_file(fitsfile,C.READWRITE)
        else:
           if not path.exists(fitsfile): raise "%s does not exist!" % (fitsfile)
           raise "Error with %s." % (fitsfile)
        self.fp=fp
        self.file=C.fitsfile(self.fp)
        self.rawkeys = []
        self.rawkeys_hdu = -99

  def __getitem__(self,key):
        if upper(key) == "NAXES": return self.naxes()
        elif key.upper() in self.keys(): return self.getval(key)
        else: return "N/A"

  def __setitem__(self,key,val):
       status=self.modkey(key,val)

  def __delitem__(self,key):
       status=C.delete_key(self.file,key)

  def newhead(self,simplev,bitpixv,naxisv,naxesv):
        status=C.create_img(self.file,bitpixv,naxisv,naxesv)

  def fitsfile(self):
        return self.file

  def name(self):
        return self.file.Fptr.filename

  def mode(self):
        return self.file.Fptr.writemode

  def resize(self,naxes):
      for n in range(self.naxis()): self.delkey("naxis"+str(n+1))
      self.modkey("NAXIS",len(naxes))
      for n in range(len(naxes)):
         if n == 0: self.inskey("NAXIS","NAXIS"+str(n+1),naxes[n])
         else: self.inskey("NAXIS"+str(n),"NAXIS"+str(n+1),naxes[n])
      return C.resize_img(self.file,self.bitpix(),len(naxes),naxes)

  def reopen(self,readonly=None):
       name = self.name()
       if readonly == None: readonly=self.mode()
       elif readonly == 0: readonly=C.READWRITE
       else: readonly=C.READONLY
       self.close()
       status,fp=C.open_file(name,C.READONLY)
       self.file=C.fitsfile(fp)

  def flush(self):
      C.flush_file(self.file)

  def close(self):
      self.flush()
      C.close_file(self.file)

  def addhdu(self):
     return C.create_hdu(self.file)

  def numhdus(self):
     sn = C.get_num_hdus(self.file)
     return sn

  def goto(self,hdu):
     return C.movabs_hdu(self.file,hdu)

  def header(self,frame=1):
#     C.movabs_hdu(self.file,frame)
     header=[]; i=1
     status,rec=C.read_record(self.file,i)
     while status==0 and strip(rec) != 'END':
        header.append( rec); i+=1
        status,rec=C.read_record(self.file,i)
     return header

  def copyheader(self, o, hdu=1):
      o.goto(hdu)
      oh=o.header()
      for h in oh: C.write_record(self.file,h)

  def nkeys(self):
        return len(self.header())

  def raw_keys(self):
      cur = C.get_hdu_num(self.file)
      if cur != self.rawkeys_hdu or not self.rawkeys:
         n=self.nkeys()
         r=arange(n)+1
         f=self.file
         self.rawkeys = [C.read_keyn(f,n)[1] for n in r]
         self.rawkeys_hdu = C.get_hdu_num(self.file)
      return self.rawkeys

  def keys(self):
      keys = self.raw_keys()
      for i in range( keys.count('')): keys.remove('')
      return keys

  def empty(self):
      return nonzero(map(lambda k: (k == ''), self.raw_keys()))

  def not_empty(self):
      return nonzero(map(lambda k: (k != ''), self.raw_keys()))

  def readval(self,value):
     t = None
     if value != "T" and value != "":
       t = C.get_keytype(value)[1][0]
       ts = ["C","L","I","F","X"]
       fs = [eval,lambda c: ["F","T"].index(c),int,float,complex]
       value = fs[ts.index(t)](value)
     return value,t

  def getval(self,key,comment=0):
      value = "N/A"
      commstr = ""
      if key.upper() in self.keys():
         status,value,commstr=C.read_keyword(self.file,key)
         value = self.readval(value)[0]
      if comment: return value, commstr
      else: return value

  def vals(self):
      return map(self.getval, self.keys())

  def bitpix(self):
        return C.read_key(self.file,C.TLONG,"bitpix")[1]

  def bscale(self):
        if 'BSCALE' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"bscale")[1]
        else: return 1

  def bzero(self):
        if 'BZERO' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"bzero")[1]
        else: return None

  def crval1(self):
        if 'CRVAL1' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"crval1")[1]
        else:
           return 0.0

  def crpix1(self):
        if 'CRPIX1' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"crpix1")[1]
        else: return 1.0

  def cdelt1(self):
        if 'CDELT1' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"cdelt1")[1]
        else: return 1.0

  def crval2(self):
        if 'CRVAL2' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"crval2")[1]
        else: return 0.0

  def crpix2(self):
        if 'CRPIX2' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"crpix2")[1]
        else: return 1.0

  def cdelt2(self):
        if 'CDELT2' in self.raw_keys():
           return C.read_key(self.file,C.TDOUBLE,"cdelt2")[1]
        else: return 1.0

  def dcflag(self):
        if 'DC-FLAG' in self.raw_keys():
           return C.read_key(self.file,C.TLONG,"dc-flag")[1]
        else: return None

  def naxis(self):
        return C.read_key(self.file,C.TLONG,"naxis")[1]

  def naxisn(self,n):
        return C.read_key(self.file,C.TLONG,"naxis"+str(n))[1]

  def naxes(self,value=None):
        if value:
           for i in range(len(value)):
               C.update_key(self.file,C.TLONG,"naxis"+str(i+1),value[i])
        else:
           n=[]
           for i in range(self.naxis()): n.append(self.naxisn(i+1))
           return n

  def blank(self,val=0):
     d=zeros( self.naxes(), datatypes[bitpixes.index(self.bitpix())])
     if val: d = add(d,val)
     self.putdata(d)

  def wfunc(self,d):
     return wfunctions[datatypes.index(d.typecode())]

  def addcomment(self,comment):
     return C.write_comment(self.file,comment)

  def addhistory(self,comment):
     return C.write_history(self.file,comment)

  def putdata(self,d,group=0):
     ll = self.naxis()*[1]
     ur = self.naxes()
     self.flush()
     status=self.wfunc(d)(self.file,group+1,self.naxis(),self.naxes(),ll,ur,d)

  def delkey(self,key):
        return C.delete_key(self.file,key.upper())

  def inskey(self,prevkey,key,val,comment=''):
        if type(val) == BooleanType: ctype=C.TLOGICAL
        if type(val) == IntType: ctype=C.TINT
        if type(val) == FloatType: ctype=C.TDOUBLE
        if type(val) == LongType: ctype=C.TLONG
        if type(val) == StringType: ctype=C.TSTRING
        crap=C.read_key(self.file,C.TSTRING,prevkey)
        return C.insert_key(self.file,ctype,key.upper(),val,comment)

  def modkey(self,key,val,comment=''):
        if type(val) == BooleanType: ctype=C.TLOGICAL
        if type(val) == IntType: ctype=C.TINT
        if type(val) == FloatType: ctype=C.TDOUBLE
        if type(val) == LongType: ctype=C.TLONG
        if type(val) == StringType: ctype=C.TSTRING
        return C.update_key(self.file,ctype,key.upper(),val,comment)

  def replhead(self,h):
     lh = len(self.header())
     if len(h) < lh:
        for j in range(len(h)): C.modify_record(self.file,j+1,h[j])
        for j in range(len(h),lh): C.delete_record(self.file,j+1)
     else:
        for j in range(lh): C.modify_record(self.file,j+1,h[j])
        for j in range(lh,len(h)): C.insert_record(self.file,j+1,h[j])

  def oreplhead(self,h):
    j=len(self.header()); k=len(h); l=min((j,k))
    for i in range(l):
        C.modify_record(self.file,i+1,h[i])
    for i in range(k-j):
        C.insert_record(self.file,i+j,h[i])
    for i in range(j-k):
        C.delete_record(self.file,j-i)

  def rfunc(self):
     return rfunctions[bitpixes.index(self.bitpix())]

  def data(self,oneline=0,vnul=0.0):
     shl=self.naxes(); shl.reverse()
     ty = datatypes[ bitpixes.index(self.bitpix())]
     vnul = array([vnul]).astype(ty)[0]
     if oneline and len(shl) == 2:
        d = asarray([self.row(j) for j in range(shl[0])]).astype(ty)
     else:
        status,d=self.rfunc()(self.file,1,self.naxis(),self.naxes(),self.naxis()*[1],self.naxes(),self.naxis()*[1],vnul)
     return d

  def row(self,row,group=0,vnul=0.0):
     row += 1
     ll = [1,int(row)]
     ur = [self.naxes()[0],int(row)]
     ty = datatypes[ bitpixes.index(self.bitpix())]
     vnul = array([vnul]).astype(ty)[0]
     status,d=self.rfunc()(self.file,group+1,self.naxis(),self.naxes(),ll,ur,self.naxis()*[1],vnul)
     d = squeeze(d)
     if self.bitpix() == -32: d=d.astype(Float32)
     return d

  def col(self,col,group=0,vnul=0.0):
     col += 1
     ll = [int(col),1]
     ur = [int(col),self.naxes()[1]]
     ty = datatypes[ bitpixes.index(self.bitpix())]
     vnul = array([vnul]).astype(ty)[0]
     status,d=self.rfunc()(self.file,group+1,self.naxis(),self.naxes(),ll,ur,self.naxis()*[1],vnul)
     if self.bitpix() == -32: d=d.astype(Float32)
     return d

  def readsec(self,ll,ur,group=0,vnul=0.0):
     ll=map(lambda l: l+1,ll)
     ur=map(lambda l: l+1,ur)
     ty = datatypes[ bitpixes.index(self.bitpix())]
     vnul = array([vnul]).astype(ty)[0]
     status,d=self.rfunc()(self.file,group+1,self.naxis(),self.naxes(),ll,ur,self.naxis()*[1],vnul)
     if self.bitpix() == -32: d=d.astype(Float32)
     return d

  def readframe(self,frame,head=None,vnul=0.0):
     C.movabs_hdu(self.file,1)
     ty = datatypes[ bitpixes.index(self.bitpix())]
     vnul = array([vnul]).astype(ty)[0]
     if len(self.naxes()) == 3:
        C.movabs_hdu(self.file,1)
        ll = [1,1,int(frame)+1]
        ur = [self.naxes()[0],self.naxes()[1],int(frame)+1]
        status,d=self.rfunc()(self.file,1,self.naxis(),self.naxes(),ll,ur,self.naxis()*[1],vnul)
        d = squeeze(d)
        if self.bitpix() == -32: d=d.astype(Float32)
     else:
        C.movabs_hdu(self.file,frame+1)
        ll = [1,1]
        ur = [self.naxes()[0],self.naxes()[1]]
        status,d=self.rfunc()(self.file,1,self.naxis(),self.naxes(),ll,ur,self.naxis()*[1],vnul)
        d = squeeze(d)
        if self.bitpix() == -32: d=d.astype(Float32)

     return d

  def readrow(self,row,head=None): return self.row(row)

  def readcol(self,col): return self.col(col)

  def writesec(self,d,ll,ur,group=0):
     ll=map(lambda l: l+1,ll)
     ur=map(lambda l: l+1,ur)
     status=self.wfunc(d)(self.file,group+1,self.naxis(),self.naxes(),ll,ur,d)
     self.flush()

  def writerow(self,d,row,group=0):
     ll = [1,int(row)+1]
     ur = [self.getval("naxis1"),int(row)+1]
     status=self.wfunc(d)(self.file,group+1,self.naxis(),self.naxes(),ll,ur,d)
     self.flush()

  def writecol(self,d,col,group=0):
     ll = [int(col)+1,1]
     ur = [int(col)+1,self.getval("naxis2")]
     status=self.wfunc(d)(self.file,group+1,self.naxis(),self.naxes(),ll,ur,d)
     self.flush()

############################################################

def readfits( name, oneline=0,frame=None):

        f=FITS(name,readonly=1)
        if frame == None:
          fdata = f.data(oneline)
          fnaxis = f.naxis()
          fnaxes = f.naxes()
        else:
           if frame > f.naxes()[2]-1:
             fdata = f.data(oneline)
             fnaxis = f.naxis()
             fnaxes = f.naxes()
           else:
             fdata=f.readframe(frame)
             fnaxis = f.naxis()-1
             fnaxes = f.naxes()[:2]
        rdict={'naxis':fnaxis,'naxes':fnaxes, \
        'crval1':f.crval1(), 'crpix1':f.crpix1(), 'cdelt1':f.cdelt1(), \
        'crval2':f.crval2(), 'crpix2':f.crpix2(), 'cdelt2':f.cdelt2(), \
        'dcflag':f.dcflag(),'input':name,'bitpix':f.bitpix(), \
        'header':f.header(),'data':fdata,\
        'keyval':[f.raw_keys(),f.vals(),f.keys()]}
        f.close()
        return rdict

def parsehead( kv, k):
        r=kv[0]; v=kv[1]
        if upper(k) in r: d = v[r.index(upper(k))]
        else: d="N/A"
        return d

def qread(n):
   f=FITS(n,readonly=1)
   d=f.data()
   f.close()
   return d

def qdump(n,a,t=None,verb=1,extras=[]):
    a= asarray(a)
    s=list(a.shape)
    if verb: print "Dumping",n,s
    naxis=len(s)
    naxes=s
    naxes.reverse()
    f=FITS(n,create=1,template=t)
    if not t: f.newhead("T",-32,naxis,naxes)
    else: f.modkey("bitpix",-32); f.resize(naxes)
    f.putdata(a)
    [f.modkey(e[0],e[1]) for e in extras]
    f.close()


