''' FIT.PY

This module provides a function that allows one to fit a function to data using
the  Levenberg-Marquardt algorithm proviced by scipy (so you need that installed).
It is simply a nice wrapper around the leastsq function in scipy.

First, define a function:    f(x, y, dy, a1, a2, ..., an)

The first argument must be the x-values, the second the y-values and the third
the uncertainties in y (set to 1's if you don't have errors).  The variables
a1, a2, ...  are the parameters.  The function should return the model (NOT the
residuals, as in leastsq).

To fit the model to the data, just use the fit function:

   param, e_param, C, leastsq_out = fit(f, x, y, dy, vary={}, fixed={})

The f, x, y and dy are pretty clear.  The two dictionaries that follow define what
parameters are kept constant (fixed) and which are solved for (vary).  Just use
the parameter name (as it appears in the definition of f) as the key and either
an initial guess (in the case of vary) or the fixed value (in the case of fixed).

The output param and e_param are both dictionaries.  Output has both the fixed and
varied parameters and their final values.  e_param has the uncertainties in these
varied parameters or 0.0 if fixed.  C is the covariance matrix and leastsq_out is
the full output of the leastsq call. '''

import scipy
from scipy import linalg
from scipy import optimize
import Numeric as num

def fit(func, x, y, dy, vary, fixed):
   '''Use the Lavenberg-Marquardt method to fit func to the data (x,y +/- dy).  
   The function is of the form func(x,y,dy,x1,x2,...,xn,a1,a2,...,an) where the
   are variables to be varied and ai are to be held fixed.  However, the varied
   and fixed variables can appear in any order.  All you need to do is supply fit
   with two dictionaries:  vary = {'x1':x1, ..., 'xn':xn} and 
   fixed = {"a1":a1, ..., "an":an}.  The key identifies the argument, the value
   is either the initial guess (in the case of vary) or the fixed value (in the
   case of fixed).
   
   The function returns the followin tuple:
      
      (out, e_out, C, leastsq_out)
          out is a dictionary with  out[variable]=value solutions, including the
             variables held fixed (for easy function evaluation later on).
          e_vary is a dictionary of errors for the variables (=0 for fixed)
          C is the full covariance matrix
          var_list is a list of variables, in the order they appear in C
          leastsq_out is the full out put of optimize.leastsq (itself a tuple).'''

   vary_names = vary.keys()
   vary_names.sort()          # to keep a consistent ordering
   fixed_names = fixed.keys()
   fixed_names.sort()
   args = {}
   list = []
   tuple = []
   for name in vary_names:
      args[name] = vary[name]
      list.append(vary[name])
   for name in fixed_names:
      args[name] = fixed[name]
   tuple = (x,y,dy,args,vary_names,func)
   print args

   res = optimize.leastsq(wrap_func, list, tuple, full_output=1)

   # Now compute the covariance matrix
   ipvt = res[1]['ipvt']
   n = len(ipvt)
   fjac = res[1]['fjac']
   perm = num.take(scipy.eye(n), ipvt -1)
   r = linalg.triu(num.transpose(fjac)[:n,:])
   R = scipy.dot(r,perm)
   C = linalg.inv(scipy.dot(num.transpose(R),R))

   output = {}
   errors = {}
   i = 0
   for name in vary_names:
      output[name] = res[0][i]
      errors[name] = num.sqrt(C[i][i])
      i = i + 1
   for name in fixed:
      output[name] = fixed[name]
      errors[name] = 0.0

   return((output, errors, C, vary_names, res))
   
def wrap_func(list, x,y,dy,args,vary_names,func):
   i = 0
   for name in vary_names:
      args[name] = list[i]
      i = i + 1
   return((y - func(x,**args))/dy)

