# Module: peakfit
__version__ = "0.1"
__author__ = "Daniel K. Magee <magee@ucolick.org>"
# Fits a Gaussian, Lorentzian or Moffat model to 1D or 2D numeric array
#
# Usage: fit1d = peakfit1d(x, y, function="function")
#        fit2d = peakfit2d(x, y, z, function="function")
#        
#             Returns a best fitting model of function
#
# 1D Functions:
#
#     Gaussian    P(0)*exp(-0.5*u)
#
#     Lorentzian  P(0)/(u + 1)
#
#     Moffat      P(0)/(u + 1)**P(3)
#
#     where u = ((x - P(1))/P(2))**2
#
#     P(0)    Peak                Peak               Peak
#     P(1)    Peak Centroid       Peak Centroid      Peak Centroid
#     P(2)    Sigma               HWHM               HWHM
#     P(3)    Constant Term*      Constant Term*     Moffat Index
#     P(4)    Linear Term*        Linear Term*       Constant Term*
#     P(5)                                           Linear Term*
#                
#     * Optional term depending on value of nterms
#
#     x = array of independent variable values
#     y = array of "measured" dependent values
#
# 2D Functions:
#
#    Gaussian    P(0)*exp(-0.5*u)+v
#
#    Lorentzian  P(0)/(u + 1)+v
#
#    where u = ((x-P(3))/P(1))**2 + ((y-P(4))/P(3))**2
#    and   v = P(6) + P(7)*x + P(8)*y
#
#    P(0)   Peak value
#    P(1)   Peak half-width (x) -- gaussian sigma or HWHM
#    P(2)   Peak half-width (y) -- gaussian sigma or HWHM
#    P(3)   Peak centroid (x)
#    P(4)   Peak centroid (y)
#    P(5)   Rotation angle (radians)
#    P(6)   Constant Term*
#    P(7)   Linear Term* (x)
#    P(8)   Linear Term* (y)
#
#    * Optional term depending on value of nterms
#
#    Moffat   P(0)/(u + 1)**P(6)+v
#
#    where u = ((x-P(3))/P(1))**2 + ((y-P(4))/P(3))**2
#    and   v = P(7) + P(8)*x + P(9)*y
#
#    P(0)   Peak value
#    P(1)   Peak half-width (x) -- gaussian sigma or HWHM
#    P(2)   Peak half-width (y) -- gaussian sigma or HWHM
#    P(3)   Peak centroid (x)
#    P(4)   Peak centroid (y)
#    P(5)   Rotation angle (radians)
#    P(6)   Moffat Index
#    P(7)   Constant Term*
#    P(8)   Linear Term* (x)
#    P(9)   Linear Term* (y)

from sys import exit
from Numeric import *
from Multipack import *

# 1d Fitting

def peakfit_u1d(x, p):
	width = abs(p[2])
	return ((x - p[1])/width)**2

# Gaussian Function

def peakfit_gauss1d(p, x, y, w):
	u = peakfit_u1d(x, p)
	if len(p) > 3:
		f = p[3]
	else:
		f = 0
	if len(p) > 4:
		f = f + p[4]*x
	return (f + p[0]*exp(-0.5*u) - y)*w

# Lorentzian Function

def peakfit_lorentz1d(p, x, y, w):
	u = peakfit_u1d(x, p)
	if len(p) > 3:
		f = p[3]
	else:
		f = 0
	if len(p) > 4:
		 f = f + p[4]*x
	return (f + p[0]/(u + 1) - y)*w

# Moffat Function

def peakfit_moffat1d(p, x, y, w):
	u = peakfit_u1d(x, p)
	if len(p) > 4:
		f = p[4]
	else:
		f = 0
	if len(p) > 5:
		f = f + p[5]*x
	return (f + p[0]/(u + 1)**p[3] - y)*w
	
# Estimate parameters

def estpar1d(x, y):
	nx = len(x)
	maxx = maximum.reduce(x)
	minx = minimum.reduce(x)
	maxy = maximum.reduce(y)
	miny = minimum.reduce(y)
	totalarea = add.reduce(y)
	aveheight = totalarea/(maxx - minx)

	ge = greater_equal(y, aveheight)
	sum_ge = add.reduce(multiply(x, ge))
	sum2_ge = add.reduce(multiply(x, ge)**2) 
	ge_spd = sum2_ge/add.reduce(ge) - (sum_ge/add.reduce(ge))**2
	le = less_equal(y, aveheight)
	sum_le = add.reduce(multiply(x, le))
	sum2_le = add.reduce(multiply(x, le)**2)
	le_spd = sum2_le/add.reduce(ge) - (sum_le/add.reduce(le))**2

	if ge_spd < le_spd:
		center = x[argmax(y)]
		peak = maxy - aveheight
	else:
		center = x[argmin(y)]
		peak = miny - aveheight
	peakarea = totalarea - add.reduce(multiply(y, le))
	width = peakarea / (2*abs(peak))
	background = aveheight
	slope = (y[-1] - y[0])/nx
	return [peak, center, width, background, slope]



def peakfit1d(x, y, function="gaussian", nterms=5, yerrors=0, weighting="none", estpars="yes"):
	# Compute weighting factors
	if yerrors == 0:
		if weighting == "none":
			weights = ones(len(x))
		elif weighting == "poisson":
			weights = abs(divide(1, y))
		else:
			print "WARNING: No weighting used."
			print "You must provide YERROR array for normal weighting."
			weights = ones(len(x))
	else:
		if weighting == "none":
			weights = yerrors
		elif weighting == "poisson":
			weights = abs(divide(1, yerrors))
		else:
			weights = abs(divide(1, yerrors**2))

	if estpars != "yes":
		pars = estpars
	else:
		pars = estpar1d(x, y)

	if function == "gaussian":
		pars = pars[:nterms]
		fit1d = leastsq(peakfit_gauss1d, pars, args=(x,y,weights))[0].tolist()
	
	elif function == "lorentz":
		pars = pars[:nterms]
		fit1d = leastsq(peakfit_lorentz1d, pars, args=(x,y,weights))[0].tolist()
	
	else:
		pars = (pars[:3] + [1] + pars[3:])[:nterms]
		fit1d = leastsq(peakfit_moffat1d, pars, args=(x,y,weights))[0].tolist()

	return fit1d

# 2D Fitting

def peakfit_u2d(x, y, p):
	xc = x - p[1]
	yc = y - p[2]
	wx = abs(p[3])
	wy = abs(p[4])
	theta = p[5]
	c = cos(theta)
	s = sin(theta)
	return ((xc * (c/wx) - yc * (s/wx))**2 + (xc * (s/wy) + yc * (c/wy))**2)

# Gaussian 2d Function

def peakfit_gauss2d(p, x, y, d, w, resid=1):
	try: u = peakfit_u2d(x, y, p)
        except: return 1e18
	if len(p) > 6:
		f = p[6]
	else:
		f = 0
	if len(p) > 7:
		f = f + p[7]*x + p[8]*y
	try:
		f = f + p[0]*exp(-0.5*u)
		if resid: f = (f - d)*w
	except: f = 1e18
	return f

# Lorentzian 2d Function

def peakfit_lorentz2d(p, x, y, d, w, resid=1):
	u = peakfit_u2d(x, y, p)
	if len(p) > 6:
		f = p[6]
	else:
		f = 0
	if len(p) > 7:
		 f = f + p[7]*x + p[8]*y
	f = f + p[0]/(u + 1)
	if resid: f = (f - d)*w
	return f

# Moffat 2d Function

def peakfit_moffat2d(p, x, y, d, w):
	u = peakfit_u2d(x, y, p)
	if len(p) > 7:
		f = p[7]
	else:
		f = 0
	if len(p) > 8:
		f = f + p[8]*x + p[9]*y
	f = f + p[0]/(u + 1)**p[6]
	return (f - d)*w
	
# Estimate 2d parameters

def estpar2d(x, y, d):
	xpeakpos = argmax(add.reduce(transpose(d)))
	ypeakpos = argmax(add.reduce(d))
	xdat = d[:,ypeakpos]
	ydat = d[xpeakpos,:]
	xp = estpar1d(x, xdat)
	yp = estpar1d(y, ydat)
	peak = (xp[0]+yp[0])/2
	background = (xp[3]+yp[3])/2
	theta = 0.5
	return [peak, xp[1], yp[1], xp[2], yp[2], theta, background, xp[4], yp[4]]

def peakfit2d(x, y, d, function="gaussian", nterms=8, zerrors=None, weighting="none", estpars="yes"):
	# Compute weighting factors
	if zerrors == None:
		if weighting == "none":
			weights = ones(d.shape)
		elif weighting == "poisson":
			weights = abs(divide(1, d))
		else:
			print "WARNING: No weighting used."
			print "You must provide YERROR array for normal weighting."
			weights = ones(d.shape)
	else:
		if weighting == "none":
			weights = zerrors
		elif weighting == "poisson":
			weights = abs(divide(1, zerrors))
		else:
			weights = abs(divide(1, zerrors**2))

	if estpars != "yes":
		pars = estpars
	else:
		pars = estpar2d(x, y, d)
	
#	x = resize(x,(len(y),len(x)))
#	y = transpose(resize(y,(len(x),len(y))))
#	y = resize(y,(len(y),len(x)))

	if function == "gaussian":
		x = ravel(x)
		y = ravel(y)
		d = ravel(d)
		weights = ravel(weights)
		pars = pars[:nterms]
		fit2d = leastsq(peakfit_gauss2d, pars, args=(x,y,d,weights))
		fit2d = fit2d[0].tolist()
		fit2d[5] = (fit2d[5]  % 2*pi)
	
	elif function == "lorentz":
		x = ravel(x)
		y = ravel(y)
		d = ravel(d)
		weights = ravel(weights)
		pars = pars[:nterms]
		fit2d = leastsq(peakfit_lorentz2d, pars, args=(x,y,d,weights))[0].tolist()
	
	else:
		x = ravel(x)
		y = ravel(y)
		d = ravel(d)
		weights = ravel(weights)
		pars = (pars[:6] + [1] + pars[6:])[:nterms]
		fit2d = leastsq(peakfit_moffat2d, pars, args=(x,y,d,weights))[0].tolist()

	return fit2d


if __name__ == "__main__":
	from RandomArray import random
	npts = 1000
	x = arange(npts, typecode='d')
	params = [1000.0, 500.0, 100.0, 100.0, 0.25]
	
	yi = peakfit_gauss1d(params, x, 0.0, 1.0 )
	y = yi + random(x.shape)*sqrt(yi)
	fit1d = peakfit1d(x, y, function="gaussian",  nterms=5, weighting="poisson")
	print "Gaussian"
	print params
	print fit1d
	
	yi = peakfit_lorentz1d(params, x, 0.0, 1.0 )
	y = yi + random(x.shape)*sqrt(yi)
	fit1d = peakfit1d(x, y, function="lorentz",  nterms=5, weighting="poisson")
	print "Lorentzian"
	print params
	print fit1d
	
	params = [1000.0, 500.0, 100.0, 2.0, 100.0, 0.25]
	yi = peakfit_moffat1d(params, x, 0.0, 1.0 )
	y = yi + random(x.shape)*sqrt(yi)
	fit1d = peakfit1d(x, y, function="moffat",  nterms=6, weighting="poisson")
	print "Moffat"
	print params
	print fit1d

	xnpts = 100
	ynpts = 100
	x = arange(xnpts)
	y = arange(ynpts)
	xx = resize(x,(len(y),len(x)))
	yy = transpose(resize(y,(len(x),len(y))))
	params = [100.0, 50.0, 52.0, 10.0, 12.0, 0.61, 10.0, 0.25, 0.35,]
	zi = peakfit_gauss2d(params, xx, yy, 0.0, 1.0)
	z = zi + random((len(x),len(y)))*sqrt(zi)
	fit2d = peakfit2d(x, y, z, function="gaussian",  nterms=9, weighting="poisson")
	print "2d Gaussian"
	print params
	print fit2d


